__title__ = 'PyGeoDiff'
__description__ = 'Diff tool for geo-spatial data'
__url__ = 'https://github.com/lutraconsulting/geodiff'
__version__ = '0.3.0'
__author__ = 'Peter Petrik'
__author_email__ = 'zilolv@gmail.com'
__maintainer__ = 'Peter Petrik'
__license__ = 'MIT'
__copyright__ = '(c) 2019 Peter Petrik'